/*
 *  soapUI, copyright (C) 2004-2009 eviware.com 
 *
 *  soapUI is free software; you can redistribute it and/or modify it under the 
 *  terms of version 2.1 of the GNU Lesser General Public License as published by 
 *  the Free Software Foundation.
 *
 *  soapUI is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
 *  even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 *  See the GNU Lesser General Public License for more details at gnu.org.
 */

package com.eviware.soapui.impl.wsdl.teststeps;

import com.eviware.soapui.impl.wsdl.WsdlOperation;
import com.eviware.soapui.impl.wsdl.submit.WsdlMessageExchange;
import com.eviware.soapui.impl.wsdl.submit.transports.http.WsdlResponse;
import com.eviware.soapui.impl.wsdl.support.assertions.AssertedXPathsContainer;
import com.eviware.soapui.impl.wsdl.support.soap.SoapVersion;
import com.eviware.soapui.impl.wsdl.teststeps.actions.ShowMessageExchangeAction;
import com.eviware.soapui.model.ModelItem;
import com.eviware.soapui.model.iface.Attachment;
import com.eviware.soapui.model.iface.MessageExchange;
import com.eviware.soapui.model.testsuite.AssertedXPath;
import com.eviware.soapui.model.testsuite.MessageExchangeTestStepResult;
import com.eviware.soapui.model.testsuite.ResponseAssertedMessageExchange;
import com.eviware.soapui.support.action.swing.ActionList;
import com.eviware.soapui.support.types.StringToStringMap;
import com.eviware.soapui.support.xml.XmlUtils;

import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.List;
import java.util.Vector;

/**
 * TestStepResult for a WsdlTestRequestStep
 *
 * @author ole.matzura
 */

public class WsdlTestRequestStepResult extends WsdlTestStepResult implements
        ResponseAssertedMessageExchange, AssertedXPathsContainer, MessageExchangeTestStepResult, WsdlMessageExchange
{
   private String requestContent;
   private WsdlResponse response;
   private String domain;
   private String username;
   private String endpoint;
   private String encoding;
   private String password;
   private StringToStringMap properties;
   private boolean addedAction;
   private List<AssertedXPath> assertedXPaths;

   public WsdlTestRequestStepResult( WsdlTestRequestStep step )
   {
      super( step );
   }

   public WsdlOperation getOperation()
   {
      WsdlTestRequestStep testStep = (WsdlTestRequestStep) getTestStep();
      return testStep.getOperation();
   }

   public SoapVersion getSoapVersion()
   {
      return ((WsdlOperation) getOperation()).getInterface().getSoapVersion();
   }

   public ModelItem getModelItem()
   {
      if( response != null )
         return response.getRequest();
      else
         return null;
   }

   public String getRequestContent()
   {
      if( isDiscarded() )
         return "<discarded>";

      return requestContent;
   }

   public void setRequestContent( String requestContent )
   {
      this.requestContent = requestContent;
   }

   public WsdlResponse getResponse()
   {
      return response;
   }

   @Override
   public ActionList getActions()
   {
      if( !addedAction )
      {
         addAction( new ShowMessageExchangeAction( this, "TestStep" ), true );
         addedAction = true;
      }

      return super.getActions();
   }

   public void setResponse( WsdlResponse response )
   {
      this.response = response;
   }

   public String getDomain()
   {
      return domain;
   }

   public void setDomain( String domain )
   {
      this.domain = domain;
      addProperty( "domain", domain );
   }

   private void addProperty( String key, String value )
   {
      if( properties == null )
         properties = new StringToStringMap();

      properties.put( key, value );
   }

   public String getEncoding()
   {
      return encoding;
   }

   public void setEncoding( String encoding )
   {
      this.encoding = encoding;
      addProperty( "encoding", encoding );
   }

   public String getEndpoint()
   {
      return endpoint;
   }

   public void setEndpoint( String endpoint )
   {
      this.endpoint = endpoint;
      addProperty( "endpoint", endpoint );
   }

   public String getPassword()
   {
      return password;
   }

   public void setPassword( String password )
   {
      this.password = password;
      addProperty( "password", password );
   }

   public String getUsername()
   {
      return username;
   }

   public void setUsername( String username )
   {
      this.username = username;
      addProperty( "username", username );
   }

   public void discard()
   {
      super.discard();

      requestContent = null;
      response = null;
      properties = null;
      assertedXPaths = null;
   }

   public void writeTo( PrintWriter writer )
   {
      super.writeTo( writer );

      writer.println( "----------------------------------------------------" );
      writer.println( "Encoding: " + getEncoding() );
      writer.println( "Endpoint: " + getEndpoint() );
      writer.println( "Username: " + getUsername() );
      writer.println( "Password: " + getPassword() );
      writer.println( "Domain: " + getDomain() );

      writer.println( "---------------- Request ---------------------------" );
      if( requestContent != null )
         writer.println( XmlUtils.prettyPrintXml( requestContent ) );
      else
         writer.println( "- missing request / garbage collected -" );

      if( response != null )
      {
         writer.println( "Request Headers: " + response.getRequestHeaders().toString() );
      }

      writer.println( "---------------- Response --------------------------" );
      if( response != null )
      {
         String respContent = response.getContentAsString();
         if( respContent != null )
            writer.println( XmlUtils.prettyPrintXml( respContent ) );

         writer.println( "Response Headers: " + response.getResponseHeaders().toString() );
      }
      else
         writer.println( "- missing response / garbage collected -" );
   }

   public StringToStringMap getProperties()
   {
      return properties;
   }

   public String getProperty( String name )
   {
      return properties.get( name );
   }

   public Attachment[] getRequestAttachments()
   {
      if( response == null || response.getRequest() == null )
         return new Attachment[0];

      return response.getRequest().getAttachments();
   }

   public StringToStringMap getRequestHeaders()
   {
      if( response == null )
         return null;

      return response.getRequestHeaders();
   }

   public Attachment[] getResponseAttachments()
   {
      if( response == null )
         return new Attachment[0];

      return response.getAttachments();
   }

   public String getResponseContent()
   {
      if( isDiscarded() )
         return "<discarded>";

      if( response == null )
         return "<missing response>";

      return response.getContentAsString();
   }

   public String getRequestContentAsXml()
   {
      return XmlUtils.seemsToBeXml( requestContent ) ? requestContent : "<not-xml/>";
   }

   public String getResponseContentAsXml()
   {
      String responseContent = getResponseContent();
      return XmlUtils.seemsToBeXml( responseContent ) ? responseContent : null;
   }

   public StringToStringMap getResponseHeaders()
   {
      if( response == null )
         return null;

      return response.getResponseHeaders();
   }

   public long getTimestamp()
   {
      if( isDiscarded() || response == null )
         return -1;

      return response.getTimestamp();
   }

   public AssertedXPath[] getAssertedXPathsForResponse()
   {
      return assertedXPaths == null ? new AssertedXPath[0] :
              assertedXPaths.toArray( new AssertedXPath[assertedXPaths.size()] );
   }

   public void addAssertedXPath( AssertedXPath assertedXPath )
   {
      if( assertedXPaths == null )
         assertedXPaths = new ArrayList<AssertedXPath>();

      assertedXPaths.add( assertedXPath );
   }

   public MessageExchange[] getMessageExchanges()
   {
      return new MessageExchange[] { this };
   }

   public byte[] getRawRequestData()
   {
      return response == null ? null : response.getRawRequestData();
   }

   public byte[] getRawResponseData()
   {
      return response == null ? null : response.getRawResponseData();
   }

   public Attachment[] getRequestAttachmentsForPart( String partName )
   {
      return null;
   }

   public Attachment[] getResponseAttachmentsForPart( String partName )
   {
      return null;
   }

   public boolean hasRawData()
   {
      return false;
   }

   public boolean hasRequest( boolean b )
   {
      return true;
   }

   public boolean hasResponse()
   {
      return response != null;
   }

   public Vector<?> getRequestWssResult()
   {
      return null;
   }

   public Vector<?> getResponseWssResult()
   {
      return response == null ? null : response.getWssResult();
   }

   public int getResponseStatusCode()
   {
      return response == null ? 0 : response.getStatusCode();
   }

   public String getResponseContentType()
   {
      return response == null ? null : response.getContentType();
   }
}